(* Title: extends Isabelle/src/Pure/term.ML
   Author: Walther Neuper 1999, Mathias Lehnfeld
   (c) due to copyright terms
*)
infix contains_one_of

(* TERM_C extends Isabelle's naming conventions: "C" indicates Isac add-ons to an Isabelle module *)
signature TERM_ISAC =
sig
  type as_string
  val empty: term
  val typ_empty: typ

  datatype lrd = D | L | R
  type path
  val string_of_path: path -> string
  val sub_at: path -> term -> term
  val go_up: path -> term -> term

  val perm: term -> term -> bool (*old original Isabelle code*)
  val contains_Var: term -> bool
  val dest_binop_typ: typ -> typ * typ * typ
  val dest_equals: term -> term * term
  val free2str: term -> string
  val ids2str: term -> string list
  val ins_concl: term -> term -> term
  val inst_abs: term -> term
  val inst_bdv: LibraryC.subst -> term -> term

  val lhs: term -> term
  val rhs: term -> term

  val mk_frac: typ -> int * (int * int) -> term
  val numerals_to_Free: term -> term
  val term_of_num: typ -> int -> term
  val num_of_term: term -> int
  val to_string: term -> string
  val int_of_str: string -> int
  val isastr_of_int: int -> string
  val int_opt_of_string: string -> int option

  val isalist2list: term -> term list
  val list2isalist: typ -> term list -> term
  val isapair2pair: term -> term * term (* rename to dest_pair, compare HOLogic.dest_string *)

  val is_atom: term -> bool
  val string_of_atom: term -> string
  val is_variable: term -> bool
  val is_bdv: string -> bool
  val is_bdv_subst: term -> bool
  val guess_bdv_typ: term -> typ
  val is_equality: term -> bool
  val is_expliceq: term -> bool
  val is_f_x: term -> bool
  val is_list: term -> bool
  val is_bool_list: term -> bool
  val dest_listT: typ -> typ
  val is_num: term -> bool
  val is_num': string -> bool
  val string_of_num: term -> string
  val variable_constant_pair: term * term -> bool

  val mk_add: term -> term -> term
  val mk_free: typ -> string -> term
  val mk_equality: term * term -> term
  val mk_factroot: string -> typ -> int -> int -> term
  val mk_Free: string * typ -> term
  val mk_thmid: string -> string -> string -> string
  val mk_num_op_num: typ -> typ -> string * typ -> int -> int -> term
  val mk_num_op_var: term -> string -> typ -> typ -> int -> term
  val mk_var_op_num: term -> string -> typ -> typ -> int -> term

  val matches: theory -> term -> term -> bool

  val parseNEW: Proof.context -> string -> term option(*old version to be eliminated*)
  val parseNEW': Proof.context -> string -> term      (*old version to be eliminated*)
  val parseNEW'': theory -> string -> term            (*old version to be eliminated*)
(*goal*)val parse: Proof.context -> string -> term
(*goal*)val parse_patt: theory -> string -> term
  (*for test/* *)
(*goal*)val parse_test: Proof.context -> string -> term
(*goal*)val parse_patt_test: theory -> string -> term

  val str_of_free_opt: term -> string option
  val str_of_int: int -> string
  val strip_imp_prems': term -> term option
  val subst_atomic_all: LibraryC.subst -> term -> bool * term

  val pairt: term -> term -> term
  val pairT: typ -> typ -> typ
  val raise_type_conflicts: term list -> unit
  val strip_trueprop: term -> term

  val var2free: term -> term
  val vars: term -> term list  (* recognises numerals, should replace "fun vars_of" TODOO*)
  val vars': term list -> term list
  val vars_of: term -> term list   (* deprecated TODOO: see differences in test/../termC.sml*)
  val dest_list': term -> term list
  val negates: term -> term -> bool

  val contains_one_of: thm * (string * typ) list -> bool
  val contains_Const_typeless: term list -> term -> bool
  val sym_trm : term -> term (* unused code, kept as hints to design ideas *)

  val string_of_detail: Proof.context -> term -> string

\<^isac_test>\<open>
  val mk_negative: typ -> term -> term
  val mk_Var: term -> term
  val scala_of_term: term -> string

  val atom_typ: Proof.context -> typ -> unit
  val atom_write: Proof.context -> term -> unit
  val atom_trace: Proof.context -> term -> unit

  val atom_write_detail: Proof.context -> term -> unit
  val atom_trace_detail: Proof.context -> term -> unit
\<close>
end

(**)
structure TermC(**): TERM_ISAC(**) =
struct
(**)

type as_string = UnparseC.term_as_string
val empty = UnparseC.term_empty
val typ_empty = UnparseC.typ_empty

datatype lrd = L (*t1 in "t1$t2"*)
             | R (*t2 in "t1$t2"*) | D; (*b in Abs(_,_,b*)
type path = lrd list; 
fun ldr2str L = "L"
  | ldr2str R = "R"
  | ldr2str D = "D";
fun string_of_path k = (strs2str' o (map ldr2str)) k;
(*go to a location in a term and fetch the resective sub-term*)
fun sub_at [] t = t
  | sub_at (D :: p) (Abs(_, _, body)) = sub_at p body
  | sub_at (L :: p) (t1 $ _) = sub_at p t1
  | sub_at (R :: p) (_ $ t2) = sub_at p t2
  | sub_at l t = raise TERM ("sub_at: no " ^ string_of_path l ^ " for ", [t]);
fun go_up l t =
  if length l > 1 then sub_at (drop_last l) t else raise ERROR ("go_up [] " ^ UnparseC.term t)

fun isastr_of_int i = if i >= 0 then string_of_int i else "-" ^ string_of_int (abs i)

fun matches thy tm pa = 
    (Pattern.match thy (pa, tm) (Vartab.empty, Vartab.empty); true)
    handle Pattern.MATCH => false

(** transform  typ / term to a String to be parsed by Scala after transport via libisabelle **)
fun string_of_detail ctxt t =
  let
    fun ato (Const (a, T)) n = "\n*** " ^ indent n ^ "Const (" ^ a ^ ", " ^ UnparseC.typ_in_ctxt ctxt T ^ ")"
      | ato (Free (a, T)) n = "\n*** " ^ indent n ^ "Free (" ^ a ^ ", " ^ UnparseC.typ_in_ctxt ctxt T ^ ")"
      | ato (Var ((a, i), T)) n =
        "\n*** " ^ indent n ^ "Var ((" ^ a ^ ", " ^ string_of_int i ^ "), " ^ UnparseC.typ_in_ctxt ctxt T ^ ")"
      | ato (Bound i) n = "\n*** " ^ indent n ^ "Bound " ^ string_of_int i
      | ato (Abs(a, T, body))  n = 
        "\n*** " ^ indent n ^ "Abs (" ^ a ^ ", " ^ UnparseC.typ_in_ctxt ctxt T ^ ",.." ^ ato body (n + 1)
      | ato (f $ t) n = ato f n ^ ato t (n + 1)
  in "\n*** " ^ ato t 0 ^ "\n***" end;


\<^isac_test>\<open>
fun scala_of_typ (Type (s, typs)) =
    enclose "Type(" ")" (quote s ^ ", " ^
      (typs |> map scala_of_typ |> commas |> enclose "List(" ")"))
  | scala_of_typ (TFree (s, sort)) =
    enclose "TFree(" ")" (quote s ^ ", " ^ (sort |> map quote |> commas |> enclose "List(" ")"))
  | scala_of_typ (TVar ((s, i), sort)) =
    enclose "TVar(" ")" (
      enclose "(" ")," (quote s ^ ", " ^ quote (string_of_int i)) ^ 
      (sort |> map quote |> commas |> enclose "List(" ")"))

fun scala_of_term (Const (s, T)) =
    enclose "Const (" ")" (quote s ^ ", " ^ scala_of_typ T)
  | scala_of_term (Free (s, T)) =
    enclose "Free(" ")" (quote s ^ ", " ^ scala_of_typ T)
  | scala_of_term (Var ((s, i), T)) =
    enclose "TVar(" ")" (
      enclose "(" ")," (quote s ^ ", " ^ quote (string_of_int i)) ^ 
      scala_of_typ T)
  | scala_of_term (Bound i) = enclose "Bound(" ")" (string_of_int i)
  | scala_of_term (Abs (s, T, t)) =
    enclose "Abs(" ")" (
      quote s ^ ", " ^
      scala_of_typ T ^ ", " ^
      scala_of_term t)
  | scala_of_term (t1 $ t2) =
    enclose "App(" ")" (scala_of_term t1 ^ ", " ^ scala_of_term t2)

(* see structure's bare bones.
   for Isabelle standard output compare 2017 "structure ML_PP" *)
fun atom_typ ctxt t =
  let
    fun ato n (Type (s, [])) = "\n*** " ^ indent n ^ "Type (" ^ s ^",[])"
      | ato n (Type (s, Ts)) = "\n*** " ^ indent n ^ "Type (" ^ s ^ ",[" ^ atol (n + 1) Ts
      | ato n (TFree (s, sort)) = "\n*** " ^ indent n ^ "TFree (" ^ s ^ ", " ^ strs2str' sort
      | ato n (TVar ((s, i), sort)) =
        "\n*** " ^ indent n ^ "TVar ((" ^ s ^ ", " ^ string_of_int i ^ strs2str' sort
    and atol n [] = "\n*** " ^ indent n ^ "]"
      | atol n (T :: Ts) = (ato n T ^ atol n Ts)
in tracing (ato 0 t ^ "\n") end;

local 
  fun ato (Const (a, _)) n = "\n*** " ^ indent n ^ "Const (" ^ a ^ ", _)"
	  | ato (Free (a, _)) n = "\n*** " ^ indent n ^ "Free (" ^ a ^ ", _)"
	  | ato (Var ((a, i), _)) n =
	    "\n*** " ^ indent n ^ "Var (" ^ a ^ ", " ^ string_of_int i ^ "), _)"
	  | ato (Bound i) n = "\n*** " ^ indent n ^ "Bound " ^ string_of_int i
	  | ato (Abs (a, _, body)) n = "\n*** " ^ indent n ^ "Abs(" ^ a ^ ", _" ^ ato body (n+1)
	  | ato (f $ t) n = (ato f n ^ ato t (n + 1))
in
  fun atom_write _ t = writeln ("\n*** -------------" ^ ato t 0 ^ "\n***");
  fun atom_trace _ t = tracing ("\n*** -------------" ^ ato t 0 ^ "\n***");
end;

fun atom_write_detail ctxt t = (writeln o (string_of_detail ctxt)) t;
fun atom_trace_detail ctxt t = (tracing o (string_of_detail ctxt)) t;
\<close>

(* contains the term a VAR(("*",_),_) ? *)
fun contains_Var (Abs(_,_,body)) = contains_Var body
  | contains_Var (f $ f') = contains_Var f orelse contains_Var f'
  | contains_Var (Var _) = true
  | contains_Var _ = false;

fun str_of_int n = 
  if n < 0 then "-" ^ ((string_of_int o abs) n)
  else string_of_int n;
val int_of_str = Value.parse_int;

val int_opt_of_string = ThmC_Def.int_opt_of_string
fun is_num' str = case int_opt_of_string str of SOME _ => true | NONE => false;

fun is_num (Const (\<^const_name>\<open>numeral\<close>, _) $ _) = true
  | is_num (Const (\<^const_name>\<open>uminus\<close>, _) $
    (Const (\<^const_name>\<open>numeral\<close>, _) $ _)) = true
  | is_num (Const (\<^const_name>\<open>one_class.one\<close>, _)) = true
  | is_num (Const (\<^const_name>\<open>uminus\<close>, _) $ Const (\<^const_name>\<open>one_class.one\<close>, _)) = true
  | is_num (Const (\<^const_name>\<open>zero_class.zero\<close>, _)) = true
  | is_num (Const (\<^const_name>\<open>uminus\<close>, _) $ Const (\<^const_name>\<open>zero_class.zero\<close>, _)) = true
  | is_num _ = false;

fun string_of_num n = (n |> HOLogic.dest_number |> snd |> string_of_int)

(* refer to Thm.lhs_of *)
fun lhs (Const (\<^const_name>\<open>HOL.eq\<close>,_) $ l $ _) = l
  | lhs t = raise ERROR("lhs called with (" ^ UnparseC.term t ^ ")");
fun rhs (Const (\<^const_name>\<open>HOL.eq\<close>,_) $ _ $ r) = r
  | rhs t = raise ERROR("rhs called with (" ^ UnparseC.term t ^ ")");


fun mk_negative T t = Const (\<^const_name>\<open>uminus\<close>, T --> T) $ t
fun mk_frac T (sg, (i1, i2)) =
  if sg = 1 then
    if i2 = 1 then HOLogic.mk_number T i1
    else Const (\<^const_name>\<open>divide_class.divide\<close>, T --> T --> T) $
      HOLogic.mk_number T i1 $ HOLogic.mk_number T i2
  else (*take negative*)
    if i2 = 1 then mk_negative T (HOLogic.mk_number T i1)
    else Const (\<^const_name>\<open>divide_class.divide\<close>, T --> T --> T) $
      mk_negative T (HOLogic.mk_number T i1) $ HOLogic.mk_number T i2

val numerals_to_Free = (* Makarius 100308 *)
  let
    fun dest_num t =
      (case try HOLogic.dest_number t of
        SOME (T, i) => SOME (Free (signed_string_of_int i, T))
      | NONE => NONE);
    fun to_str (Abs (x, T, b)) = Abs (x, T, to_str b)
      | to_str (t as (u1 $ u2)) =
          (case dest_num t of SOME t' => t' | NONE => to_str u1 $ to_str u2)
      | to_str t = perhaps dest_num t;
  in to_str end

val term_of_num = HOLogic.mk_number;
fun num_of_term t = t |> HOLogic.dest_number |> snd;
(* accomodate string-representation for int to term-orders *)
fun to_string t = t |> num_of_term |> signed_string_of_int

fun is_variable (t as Free _) = not (is_num t)
  | is_variable _ = false;
fun is_Free (Free _) = true | is_Free _ = false;
fun is_fun_id (Const _) = true
  | is_fun_id (Free _) = true
  | is_fun_id _ = false;
fun is_f_x (f $ x) = is_fun_id f andalso is_Free x
  | is_f_x _ = false;
(* precondition: TermC.is_atom v andalso TermC.is_atom c *)
fun is_const (Const _) = true | is_const _ = false;
fun variable_constant_pair (v, c) =
  if (is_variable v andalso (is_const c orelse is_num c)) orelse
     (is_variable c andalso (is_const v orelse is_num v))
  then true
  else false

fun vars t =
  let
    fun scan vs (Const _) = vs
      | scan vs (t as Free _) = (*if is_num' s then vs else*) t :: vs
      | scan vs (t as Var _) = t :: vs
      | scan vs (Bound _) = vs 
      | scan vs (Abs (_, _, t)) = scan vs t
      | scan vs (t1 $ t2) = (scan vs t1) @ (scan vs t2)
  in ((distinct op =) o (scan [])) t end;
fun vars' ts = ts |> map vars |> flat |> distinct op =

(* bypass Isabelle's Pretty, which requires ctxt *)
fun ids2str t =
  let
    fun scan vs (t as Const (s, _) $ arg) =
        if is_num t then vs else scan (s :: vs) arg
      | scan vs (Const (s as "Partial_Fractions.AA", _)) = s :: vs (*how get rid of spec.case?*)
      | scan vs (Const (s as "Partial_Fractions.BB", _)) = s :: vs (*how get rid of spec.case?*)
      | scan vs (Const _) = vs
      | scan vs (Free (s, _)) = if is_num' s then vs else s :: vs
      | scan vs (Var ((s, i), _)) = (s ^ "_" ^ string_of_int i) :: vs
      | scan vs (Bound _) = vs 
      | scan vs (Abs (s, _, t)) = scan (s :: vs) t
      | scan vs (t1 $ t2) = (scan vs t1) @ (scan vs t2)
  in ((distinct op =) o (scan [])) t
  end;
fun is_bdv str = case Symbol.explode str of "b"::"d"::"v"::_ => true | _ => false;
(* instantiate #prop thm with bound vars_of (as Free) *)
fun inst_bdv [] t = t
  | inst_bdv (instl: (term*term) list) t =
    let
      fun subst (v as Var((s, _), T)) = 
          (case Symbol.explode s of
            "b"::"d"::"v"::_ => if_none (assoc(instl,Free(s,T))) (Free(s,T))
          | _ => v)
        | subst (Abs(a, T, body)) = Abs(a, T, subst body)
        | subst (f $ t') = subst f $ subst t'
        | subst t = if_none (assoc (instl, t)) t
    in  subst t  end;

(* is a term a substitution for a bdv as found in programs and tactics *)
fun is_bdv_subst (Const (\<^const_name>\<open>Cons\<close>, _) $
      (Const (\<^const_name>\<open>Pair\<close>, _) $ str $ _) $ _) = is_bdv (HOLogic.dest_string str)
  | is_bdv_subst _ = false;

(* this shall be improved due to future requirements *)
fun guess_bdv_typ t = t |> vars |> hd |> type_of

fun free2str (Free (s, _)) = s
  | free2str t = raise ERROR ("free2str not for " ^ UnparseC.term t);
fun str_of_free_opt (Free (s, _)) = SOME s
  | str_of_free_opt _ = NONE

(* compare Logic.unvarify_global, which rejects Free *)
fun var2free (t as Const _) = t
  | var2free (t as Free _) = t
  | var2free (Var((s, _), T)) = Free (s,T)
  | var2free (t as Bound _) = t 
  | var2free (Abs(s, T, t)) = Abs(s, T, var2free t)
  | var2free (t1 $ t2) = (var2free t1) $ (var2free t2);
  

\<^isac_test>\<open> (*TODO: check with new numerals --vv*)
(* Logic.varify does NOT take care of 'Free ("1", _)'*)
fun mk_Var (t as Const _) = t
  | mk_Var (t as Free (s, T)) = if is_num' s then t else Var ((s, 0), T)
  | mk_Var (t as Var _) = t
  | mk_Var (t as Bound _) = t 
  | mk_Var (Abs (s, T, t)) = Abs (s, T, mk_Var t)
  | mk_Var (t1 $ t2) = (mk_Var t1) $ (mk_Var t2);
\<close>

fun mk_listT T = Type (\<^type_name>\<open>list\<close>, [T]);
fun list_const T = Const (\<^const_name>\<open>Cons\<close>, [T, mk_listT T] ---> mk_listT T);
fun list2isalist T [] = Const (\<^const_name>\<open>Nil\<close>, mk_listT T)
  | list2isalist T (t :: ts) = (list_const T) $ t $ (list2isalist T ts);

fun isapair2pair (Const (\<^const_name>\<open>Pair\<close>,_) $ a $ b) = (a, b)
  | isapair2pair t = 
    raise ERROR ("isapair2pair called with " ^ UnparseC.term t);
fun isalist2list ls =
  let
    fun get es (Const(\<^const_name>\<open>Cons\<close>, _) $ t $ ls) = get (t :: es) ls
      | get es (Const(\<^const_name>\<open>Nil\<close>, _)) = es
      | get _ t = raise TERM ("isalist2list applied to NON-list: ", [t])
  in (rev o (get [])) ls end;

fun is_list ((Const (\<^const_name>\<open>Cons\<close>, _)) $ _ $ _) = true
  | is_list _ = false;
fun dest_listT (Type (\<^type_name>\<open>list\<close>, [T])) = T
  | dest_listT T = raise TYPE ("dest_listT: list type expected", [T], []);
fun is_bool_list t =
  (if dest_listT (Term.type_of t) = HOLogic.boolT then true else false)
  handle TYPE _ => false


fun dest_binop_typ (Type (\<^type_name>\<open>fun\<close>, [range, Type (\<^type_name>\<open>fun\<close>, [arg2, arg1])])) = (arg1, arg2, range)
  | dest_binop_typ _ = raise ERROR "dest_binop_typ: not binary";
fun dest_equals (Const(\<^const_name>\<open>HOL.eq\<close>, _) $ t $ u)  =  (t, u) (* Pure/logic.ML: Const ("==", ..*)
  | dest_equals t = raise TERM ("dest_equals'", [t]);
fun is_equality (Const(\<^const_name>\<open>HOL.eq\<close>,_) $ _ $ _)  =  true  (* logic.ML: Const ("=="*)
  | is_equality _ = false;
fun mk_equality (t, u) = (Const(\<^const_name>\<open>HOL.eq\<close>, [type_of t, type_of u] ---> HOLogic.boolT) $ t $ u); 
fun is_expliceq (Const(\<^const_name>\<open>HOL.eq\<close>,_) $ (Free _) $ _)  =  true
  | is_expliceq _ = false;
fun strip_trueprop (Const (\<^const_name>\<open>Trueprop\<close>, _) $ t) = t
  | strip_trueprop t = t;

(* (A1==>...An==>B) goes to (A1==>...An==>)   Pure/logic.ML: term -> term list*)
fun strip_imp_prems' (Const (\<^const_name>\<open>Pure.imp\<close>, _) $ A $ t) = 
    let
      fun coll_prems As (Const(\<^const_name>\<open>Pure.imp\<close>, _) $ A $ t) = 
          coll_prems (As $ (Logic.implies $ A)) t
        | coll_prems As _ = SOME As
    in coll_prems (Logic.implies $ A) t end
  | strip_imp_prems' _ = NONE;  (* *)

(* (A1==>...An==>) (B) goes to (A1==>...An==>B), where B is lowest branch, 2002 Pure/thm.ML *)
fun ins_concl (Const (\<^const_name>\<open>Pure.imp\<close>, _) $ A $ t) B = Logic.implies $ A $ (ins_concl t B)
  | ins_concl (Const (\<^const_name>\<open>Pure.imp\<close>, _) $ A    ) B = Logic.implies $ A $ B
  | ins_concl t B =  raise TERM ("ins_concl", [t, B]);

fun vperm (Var _, Var _) = true  (* 2002 Pure/thm.ML *)
  | vperm (Abs (_, _, s), Abs (_, _, t)) = vperm (s, t)
  | vperm (t1 $ t2, u1 $ u2) = vperm (t1, u1) andalso vperm (t2, u2)
  | vperm (t, u) = (t = u);

(*2002 cp from Pure/term.ML --- since 2009 in Pure/old_term.ML*)
fun mem_term (_, []) = false
  | mem_term (t, t' :: ts) = t aconv t' orelse mem_term (t, ts);
fun subset_term ([], _) = true
  | subset_term (x :: xs, ys) = mem_term (x, ys) andalso subset_term (xs, ys);
fun eq_set_term (xs, ys) =
    xs = ys orelse (subset_term (xs, ys) andalso subset_term (ys, xs));
(*a total, irreflexive ordering on index names*)
fun xless ((a, i), (b, j): indexname) = i<j  orelse  (i = j andalso a < b);
(*a partial ordering (not reflexive) for atomic terms*)
fun atless (Const (a, _), Const (b, _)) = a < b
  | atless (Free (a, _), Free (b, _)) = a < b
  | atless (Var (v, _), Var (w, _)) = xless (v, w)
  | atless (Bound i, Bound j) =  i < j
  | atless _ = false;
(*insert atomic term into partially sorted list, suppressing duplicates (?)*)
fun insert_aterm (t,us) =
  let fun inserta [] = [t]
        | inserta (us as u::us') =
              if atless(t,u) then t::us
              else if t=u then us (*duplicate*)
              else u :: inserta us'
  in inserta us end;

(* Accumulates the Vars in the term, suppressing duplicates *)
fun add_term_vars (t, vars: term list) = case t of
    Var   _ => insert_aterm (t, vars)
  | Abs (_, _, body) => add_term_vars (body, vars)
  | f$t =>  add_term_vars (f, add_term_vars (t, vars))
  | _ => vars;
fun term_vars t = add_term_vars (t, []);

(*2002 Pure/thm.ML *)
fun var_perm (t, u) = vperm (t, u) andalso eq_set_term (term_vars t, term_vars u);
(*2002 fun decomp_simp, Pure/thm.ML *)
fun perm lhs rhs = var_perm (lhs, rhs) andalso not (lhs aconv rhs) andalso not (is_Var lhs);


fun pairT T1 T2 = Type (\<^type_name>\<open>prod\<close>, [T1, T2]);
fun PairT T1 T2 = ([T1, T2] ---> Type (\<^type_name>\<open>prod\<close>, [T1, T2]));
fun pairt t1 t2 = Const (\<^const_name>\<open>Pair\<close>, PairT (type_of t1) (type_of t2)) $ t1 $ t2;

fun mk_factroot op_(*=thy.sqrt*) T fact root = 
  Const (\<^const_name>\<open>times\<close>, [T, T] ---> T) $ (term_of_num T fact) $
    (Const (op_, T --> T) $ term_of_num T root);
fun mk_var_op_num v op_ optype ntyp n = Const (op_, optype) $ v $ HOLogic.mk_number ntyp n;
fun mk_num_op_var v op_ optype ntyp n = Const (op_, optype) $ HOLogic.mk_number ntyp n $ v;
fun mk_num_op_num T1 T2 (op_, Top) n1 n2 =
  Const (op_, Top) $ HOLogic.mk_number T1 n1 $ HOLogic.mk_number T2 n2;
fun mk_thmid thmid n1 n2 = 
  thmid ^ (strip_thy n1) ^ "_" ^ (strip_thy n2);
fun mk_add t1 t2 =
  let
    val (T1, T2) = (type_of t1, type_of t2)
  in
    if T1 <> T2 then raise TYPE ("mk_add gets ", [T1, T2], [t1,t2])
    else (Const (\<^const_name>\<open>plus\<close>, [T1, T2] ---> T1) $ t1 $ t2)
  end;

(** transform binary numeralsstrings **)
fun mk_Free (s,T) = Free (s, T);
fun mk_free T s =  Free (s, T);

(*Special case: one argument cp from Isabelle2002/src/Pure/term.ML*)
fun subst_bound (arg, t) =
  let
    fun subst (t as Bound i, lev) =
        if i < lev then t (*var is locally bound*)
        else if i = lev then incr_boundvars lev arg
        else Bound (i - 1) (*loose: change it*)
      | subst (Abs(a, T, body), lev) = Abs (a, T, subst (body, lev + 1))
      | subst (f$t, lev) =  subst(f, lev)  $  subst(t, lev)
      | subst (t, _) = t
  in subst (t, 0)  end;

(* instantiate let; necessary for scan_up1 *)
fun inst_abs (Const sT) = Const sT
  | inst_abs (Free sT) = Free sT
  | inst_abs (Bound n) = Bound n
  | inst_abs (Var iT) = Var iT
  | inst_abs (Const (\<^const_name>\<open>Let\<close>,T1) $ e $ (Abs (v, T2, b))) = 
    let val b' = subst_bound (Free (v, T2), b); (*fun variant_abs: term.ML*)
    in Const (\<^const_name>\<open>Let\<close>, T1) $ inst_abs e $ (Abs (v, T2, inst_abs b')) end
  | inst_abs (t1 $ t2) = inst_abs t1 $ inst_abs t2
  | inst_abs t = t;


(*** parse ***)

(*/----- old versions to be eliminated -------------------------------------------------------\*)
fun parseNEW ctxt str = \<^try>\<open>Syntax.read_term ctxt str\<close>;
fun parseNEW' ctxt str = 
  case parseNEW ctxt str of
    SOME t => t
  | NONE => raise TERM ("NO parseNEW' for " ^ str, [])
fun parseNEW'' thy str =
  case parseNEW (Proof_Context.init_global thy) str of
    SOME t => t
  | NONE => raise TERM ("NO parseNEW'' for " ^ str, [])
(*\----- old versions to be eliminated -------------------------------------------------------/*)

(* to be replaced by parse..*)
fun parse ctxt str = Syntax.read_term_global (Proof_Context.theory_of ctxt) str;
(* parse term patterns; Var ("v",_), i.e. "?v", are required for instantiation *)
fun parse_patt thy str = (thy, str)
  |>> Proof_Context.init_global
  |-> Proof_Context.read_term_pattern

(** parse in test/* **)
(*
  This bypasses building ctxt at the begin of a calculation
  and thus borrows adapt_to_type (used for where_-parsed terms from Know_Store).
*)
fun parse_test ctxt str = str 
  |> Syntax.read_term_global (Proof_Context.theory_of ctxt)
  |> Model_Pattern.adapt_term_to_type ctxt
fun parse_patt_test thy str = (thy, str)
  |>> Proof_Context.init_global
  |-> Proof_Context.read_term_pattern
  |> Model_Pattern.adapt_term_to_type (Proof_Context.init_global thy)


fun is_atom (Const _) = true
  | is_atom (Free _) = true
  | is_atom (Var _) = true
  | is_atom t = is_num t;
fun string_of_atom (t as Const (\<^const_name>\<open>numeral\<close>, _) $ _) = to_string t
  | string_of_atom (t as Const (\<^const_name>\<open>one_class.one\<close>, _)) = to_string t
  | string_of_atom (t as Const (\<^const_name>\<open>zero_class.zero\<close>, _)) = to_string t
  | string_of_atom (Const (str, _)) = str
  | string_of_atom (Free (str, _)) = str
  | string_of_atom (Var ((str, int), _)) = str ^ "_" ^ string_of_int int
  | string_of_atom _ = "DUMMY-string_of_atom";

(* from Pure/term.ML; reports if ALL Free's have found a substitution
   (required for evaluating the preconditions of _incomplete_ models) *)
fun subst_atomic_all [] t = (false (*TODO may be 'true' for some terms ?*), t)
  | subst_atomic_all instl t =
    let
      fun subst (Abs (a, T, body)) = 
          let
            val (all, body') = subst body
          in (all, Abs(a, T, body')) end
        | subst (f$tt) = 
	        let
	          val (all1, f') = subst f
	          val (all2, tt') = subst tt
	        in (all1 andalso all2, f' $ tt') end
        | subst (t as Free _) = 
	        if is_num t then (true, t) (*numerals cannot be subst*)
	        else (case assoc (instl, t) of
					  SOME t' => (true, t')
				  | NONE => (false, t))
        | subst t = (true, if_none (assoc(instl,t)) t)
    in subst t end;

fun op contains_one_of (thm, ids) =
  Term.exists_Const (fn id => member op= ids id) (Thm.prop_of thm)

fun var_for vs (t as Const (str, _)) id =
    if is_num t then vs
    else if id = strip_thy str then t :: vs else vs
  | var_for vs (t as Free (str, _)) id = if id = str then t :: vs else vs
  | var_for vs (t as Var (idn, _)) id = if id = Term.string_of_vname idn then t :: vs else vs
  | var_for vs (Bound _) _ = vs
  | var_for vs (Abs (_, _, t)) id = var_for vs t id
  | var_for vs (t1 $ t2) id = (var_for vs t1 id) @ (var_for vs t2 id)

val poly_consts = (* TODO: adopt syntax-const from Isabelle*)
  [\<^const_name>\<open>plus\<close>, \<^const_name>\<open>minus\<close>,
  \<^const_name>\<open>divide\<close>, \<^const_name>\<open>times\<close>,
  \<^const_name>\<open>realpow\<close>];
(* treat Free, Const, Var as vars_of in polynomials *)
fun vars_of t =
  let
    val var_ids = t |> ids2str |> subtract op = poly_consts |> map strip_thy |> sort string_ord
  in (map (var_for [] t) var_ids) |> flat |> distinct op = end

(* this may decompose an object-language isa-list;
   use only, if description is not available, eg. not input ?WN:14.5.03 ??!?*)
fun dest_list' t = if is_list t then isalist2list t  else [t];

fun negat (Const (\<^const_name>\<open>Not\<close>, _) $ P, P') = P = P'
  | negat _ = false
fun negates p1 p2 = negat (p1, p2) orelse negat (swap (p1, p2));

fun raise_type_conflicts ts =
  let
    val dups = duplicates (op =) (map (fst o dest_Free) ts)
    val confl = filter (fn Free (str, _) => member op = dups str
                         | _ => false) ts
  in
    if confl = []
    then ()
    else raise TYPE ("formalisation inconsistent w.r.t. type inference: ",
      map (snd o dest_Free)confl, confl)
  end

(* expects t as Const *)
fun contains_Const_typeless ts t = (t
  |> strip_comb |> fst
  |> member (fn (t1, t2) => fst (dest_Const t1) = fst (dest_Const t2)) ts
) handle TERM("dest_Const", _) => raise TERM ("contains_Const_typeless", [t])

(* WN100910 weaker than fun sym_thm for Theory.axioms_of in isa02 *)
fun sym_trm trm =
  let
    val (lhs, rhs) = (dest_equals o strip_trueprop o Logic.strip_imp_concl) trm
    val trm' = case strip_imp_prems' trm of
	      NONE => mk_equality (rhs, lhs)
	    | SOME cs => ins_concl cs (mk_equality (rhs, lhs))
  in trm' end


end
